
from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence, Iterable, Mapping
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.live_thread_ASYNC import LiveThread, LiveUpdate

from ...models.live_thread import ContributorList, Contributor
from ...model_loaders.live_thread_ASYNC import load_live_thread, load_live_update
from ...iterators.chunking import chunked
from ...iterators.call_chunk_chaining_async_iterator import CallChunkChainingAsyncIterator
from ...iterators.async_call_chunk import AsyncCallChunk
from ...util.base_conversion import to_base36
from ...pagination.paginator_chaining_async_iterator import ImpartedPaginatorChainingAsyncIterator
from ...pagination.paginators.live_thread_async1 import LiveUpdateListingAsyncPaginator
from ... import http

class LiveThreadProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client

    async def get(self, idt: str) -> Optional[LiveThread]:
        """Get a live thread.

        .. .PARAMETERS

        :param `str` idt:

        .. .RETURNS

        :rtype: `Optional`\\[:class:`~.models.live_thread_ASYNC.LiveThread`]
        """
        try:
            root = await self._client.request('GET', f'/live/{idt}/about')
        except http.exceptions.StatusCodeException as e:
            if e.status_code == 404:
                return None
            raise
        return load_live_thread(root['data'], self._client)

    def bulk_fetch(self, idts: Iterable[str]) -> CallChunkChainingAsyncIterator[LiveThread]:
        """Bulk fetch live threads.

        Entries are returned in the same order as specified.

        If one of the IDs in a batch does not exist then a 500 HTTP error is returned.

        .. .PARAMETERS

        :param `Iterable[str]` idts:

        .. .RETURNS

        :rtype: :class:`~.iterators.call_chunk_chaining_async_iterator.CallChunkChainingAsyncIterator`\\[:class:`~.models.live_thread_ASYNC.LiveThread`]

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `500`:
                A specified ID does not exist.
        """
        async def mass_fetch(idts: Sequence[str]) -> Sequence[LiveThread]:
            idts_str = ','.join(idts)
            root = await self._client.request('GET', '/api/live/by_id/' + idts_str)
            return [load_live_thread(o['data'], self._client) for o in root['data']['children']]

        return CallChunkChainingAsyncIterator(AsyncCallChunk(mass_fetch, idfs) for idfs in chunked(idts, 100))

    async def create(self,
        title: str,
        description: str = '',
        resources: str = '',
        nsfw: bool = False,
    ) -> str:
        """Create a live thread.

        .. .PARAMETERS

        :param `str` title:
            Title. A string no longer than 120 characters.
        :param `str` description:
            Description. Markdown text. This text is displayed under the title.

            Default: empty string.
        :param `str` resources:
            The resources. Markdown text. This text is displayed on the sidebar.

            Default: empty string.
        :param `bool` nsfw:
            Mark the live thread as NSFW.

            Default false.

        .. .RETURNS

        :returns: The ID of the newly created live thread.
        :rtype: `str`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_TEXT`:
                The specified title was empty.
            + `RATELIMIT`:
                You must wait one minute before creating another live thread.
        """
        def g() -> Iterable[tuple[str, str]]:
            yield ('title', title)
            yield ('description', description)
            yield ('resources', resources)
            if nsfw: yield ('nsfw', '1')

        root = await self._client.request('POST', '/api/live/create', data=dict(g()))
        return root['json']['data']['id']

    async def config(self,
        idt: str,
        title: str,
        description: str = '',
        resources: str = '',
        nsfw: bool = False,
    ) -> None:
        """Configure the live thread.

        Requires the `settings` live thread permission.

        All parameters must be specified otherwise they will be set to their effective defaults.

        .. .PARAMETERS

        :(parameters): Same as :meth:`.create`.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - There is no user context.
               - You do not have the `settings` live thread permission.
               - You do not have permission.

            + `404`:
                The specified live thread does not exist.
        """
        def g() -> Iterable[tuple[str, str]]:
            yield ('title', title)
            yield ('description', description)
            yield ('resources', resources)
            if nsfw: yield ('nsfw', '1')

        await self._client.request('POST', f'/api/live/{idt}/edit', data=dict(g()))

    async def close(self, idt: str) -> None:
        """Close a live thread.

        This permanently closes the live thread, disallowing future updates.

        .. .PARAMETERS

        :param `str` idt:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES


        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - There is no user context.
               - You do not have the `close` permission.
               - The live thread is already closed.

            + `404`:
                The specified live thread does not exist.
        """
        await self._client.request('POST', f'/api/live/{idt}/close_thread')

    async def get_update(self, idt: str, uuid: str) -> LiveUpdate:
        """Get a specific live update in a live thread.

        .. .PARAMETERS

        :param `str` idt:
        :param `str` uuid:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `404`:
                The specified live thread ID or live update ID does not exist.
        """
        root = await self._client.request('GET', f'/live/{idt}/updates/{uuid}')
        return load_live_update(root['data'], self._client)

    def pull(self, idt: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingAsyncIterator[LiveUpdateListingAsyncPaginator, LiveUpdate]:
        """Pull live updates from a live thread.

        .. .PARAMETERS

        :param `str` idt:
        :param `str` uuid:

        .. .RETURNS

        :rtype: :class:`~.pagination.paginator_chaining_async_iterator.ImpartedPaginatorChainingAsyncIterator`\\[:class:`~.pagination.paginators.live_thread_async1.LiveUpdateListingAsyncPaginator`, :class:`~.models.live_thread_ASYNC.LiveUpdate`]

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `404`:
                The specified live thread does not exist.
        """
        p = LiveUpdateListingAsyncPaginator(self._client, f'/live/{idt}')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    async def create_update(self, idt: str, body: str) -> None:
        """Post a live update to the thread.

        .. .PARAMETERS

        :param `str` idt:
        :param `str` body:
            Markdown text.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_TEXT`:
                The `body` parameter was empty.
        """
        await self._client.request('POST', f'/api/live/{idt}/update', data={'body': body})

    async def strike_update(self, idt: str, uuid: str) -> None:
        """Strike the content of a live update.

        Requires that specified update must have been authored by the user
        or that you have the `edit` permission.

        Striken updates cannot be unstriken.

        If an already striken item is striken it is treated as a success.

        .. .PARAMETERS

        :param `str` idt:
        :param `str` uuid:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_THING_ID`:
               - The live update specified by `uuid` was empty.
               - The live update specified by `uuid` does not exist.

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You don't have permission.
            + `404`:
                The live thread specified by `idt` does not exist.
        """
        await self._client.request('POST', f'/api/live/{idt}/strike_update', data={'id': 'LiveUpdate_' + uuid})

    async def delete_update(self, idt: str, uuid: str) -> None:
        """Delete a live update.

        Requires that specified update must have been authored by the current user
        or that you have the `edit` permission.

        If an already deleted update is specified, the action will be treated as a success.
        But specifying a non-existing update ID will cause a `NO_THING_ID` API error.

        .. .PARAMETERS

        :param `str` idt:
        :param `str` uuid:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_THING_ID`:
               - The live update specified by `uuid` was empty.
               - The live update specified by `uuid` does not exist.

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You don't have permission.
            + `404`:
                The live thread specified by `idt` does not exist.
        """
        await self._client.request('POST', f'/api/live/{idt}/delete_update', data={'id': 'LiveUpdate_' + uuid})

    async def list_contributors(self, idt: str) -> ContributorList:
        """Get a list of users that contribute to a thread.

        Returns an object with a `.contributors` attribute containing
        a list of contributors, and a `.invitations` attribute containing
        a list of users that have been invited to be contributors.

        The `.invitations` list will be empty if the current user does not have
        the `manage` live thread permission.

        .. .PARAMETERS

        :param `str` idt:

        .. .RETURNS

        :rtype: :class:`~.models.live_thread.ContributorList`

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `404`:
                The live thread specified by `idt` does not exist.
        """
        root = await self._client.request('GET', f'/live/{idt}/contributors')
        if isinstance(root, (dict, Mapping)):
            return ContributorList([Contributor(d) for d in root['data']['children']], ())

        contributors = [Contributor(d) for d in root[0]['data']['children']]
        invitations = [Contributor(d) for d in root[0]['data']['children']]
        return ContributorList(contributors, invitations)

    async def send_contributor_invite(self, idt: str, user: str, permissions: Iterable[str]) -> None:
        """Invite a user to contribute to the live thread.

        Requires the `manage` live thread permission.

        .. .PARAMETERS

        :param `str` idt:
        :param `str` user:
        :param `Iterable[str]` permissions:
            Values: empty string, `all`, `close`, `discussions`, `edit`, `manage`,
            `settings`, `update`.

            Default: empty string. On the interface it'll say 'no permissions'.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `NO_USER`:
                The `user` parameter was empty.
            + `USER_DOESNT_EXIST`:
                The specified user does not exist.
            + `LIVEUPDATE_ALREADY_CONTRIBUTOR`:
                The specified user is already a contributor or has already been invited.
            + `INVALID_PERMISSIONS`:
                An invalid permission was specified.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - There is no user context.
               - You do not have the `manage` permission.

            + `404`:
                The specified live thread does not exist.
        """
        data = {
            'type': 'liveupdate_contributor_invite',
            'name': user,
            'permissions': ','.join('+' + p for p in permissions),
        }
        await self._client.request('POST', f'/api/live/{idt}/invite_contributor', data=data)

    async def accept_contributor_invite(self, idt: str) -> None:
        """Accept an invitation to contribute to a live thread.

        .. .PARAMETERS

        :param `str` idt:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `LIVEUPDATE_NO_INVITE_FOUND`:
                You don't have an invitation for the specified live thread.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `404`:
                The specified live thread does not exist.
        """
        await self._client.request('POST', f'/api/live/{idt}/accept_contributor_invite')

    async def revoke_contributor_invite(self, idt: str, user_id: int) -> None:
        """Revoke an outstanding contributor invite.

        Requires the `manage` live thread permission.

        If attempting to remove the invite for a user that was not invited,
        the action is treated as a success.

        .. .PARAMETERS

        :param `str` idt:
        :param `int` user_id:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `404`:
               - There is no user context.
               - You do not have the `manage` permission.
               - You do not have permission.
        """
        id36 = to_base36(user_id)
        await self._client.request('POST', f'/api/live/{idt}/rm_contributor_invite', data={'id': 't2_' + id36})

    async def leave_contributor(self, idt: str) -> None:
        """Abdicate contributorship of the thread.

        It is possible to leave a live thread and not have any contributors to it.

        If leaving a live thread you were not a contributor to,
        the action is treated as a success.

        .. .PARAMETERS

        :param `str` idt:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `404`:
                The specified live thread does not exist.
        """
        await self._client.request('POST', f'/api/live/{idt}/leave_contributor')

    async def remove_contributor(self, idt: str, user_id: int) -> None:
        """Revoke another user's contributorship.

        Requires the `manage` live thread permission.

        It is possible to remove your own contributorship, having the same effect as
        :meth:`.leave_contributor`.

        If attempting to remove the invite for a user that was not invited,
        the action is treated as a success.

        .. .PARAMETERS

        :param `str` idt:
        :param `int` user_id:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
              - There is no user context.
              - You are not a contributor to the live thread that has the `manage` permission.

            + `404`:
                The specified live thread does not exist.
            + `500`:
                The specified user was invalid.
        """
        id36 = to_base36(user_id)
        await self._client.request('POST', f'/api/live/{idt}/rm_contributor', data={'id': 't2_' + id36})

    async def set_contributor_permissions(self, idt: str, user: str, permissions: Iterable[str]) -> None:
        """Change a contributor's permissions.

        Requires the `manage` live thread permission.

        .. .PARAMETERS

        :param `str` idt:
        :param `str` user:
        :param `Iterable[str]` permissions:
            Same as :meth:`.send_contributor_invite`.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_USER`:
                The `user` parameter was empty.
            + `USER_DOESNT_EXIST`:
                The specified user does not exist.
            + `INVALID_PERMISSIONS`:
                An invalid permission was specified.
            + `LIVEUPDATE_NOT_CONTRIBUTOR`:
                The specified user is not a contributor.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
              - There is no user context.
              - You do not have the `manage` live thread permission.

            + `404`:
                The specified live thread does not exist.
        """
        data = {
            'type': 'liveupdate_contributor',
            'name': user,
            'permissions': ','.join('+' + p for p in permissions),
        }
        await self._client.request('POST', f'/api/live/{idt}/set_contributor_permissions', data=data)

    async def set_contributor_invite_permissions(self, idt: str, user: str, permissions: Iterable[str]) -> None:
        """Change a contributor invite's permissions.

        .. .PARAMETERS

        :param `str` idt:
        :param `str` user:
        :param `Iterable[str]` permissions:
            Same as :meth:`.send_contributor_invite`.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_USER`:
                The `user` parameter was empty.
            + `USER_DOESNT_EXIST`:
                The specified user does not exist.
            + `INVALID_PERMISSIONS`:
                An invalid permission was specified.
            + `LIVEUPDATE_NO_INVITE_FOUND`:
                The specified user does not have an invite.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
              - There is no user context.
              - You do not have the `manage` live thread permission.

            + `404`:
                The specified live thread does not exist.
        """
        data = {
            'type': 'liveupdate_contributor_invite',
            'name': user,
            'permissions': ','.join('+' + p for p in permissions),
        }
        await self._client.request('POST', f'/api/live/{idt}/set_contributor_permissions', data=data)




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...types import JSON_ro

import json

class MiscProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client

    async def convert_richtext_to_markdown(self, rt: Mapping[str, JSON_ro]) -> str:
        """Convert richtext JSON to markdown text."""
        root = await self._client.request('POST', '/api/convert_rte_body_format',
                files={'output_mode': 'markdown', 'richtext_json': json.dumps(rt)})
        return root['output']

    async def convert_markdown_to_richtext(self, md: str) -> Mapping[str, JSON_ro]:
        """Convert markdown text to richtext JSON."""
        root = await self._client.request('POST', '/api/convert_rte_body_format',
                files={'output_mode': 'rtjson', 'markdown_text': md})
        return root['output']




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Mapping, Union, \
        IO, Sequence, ClassVar, cast
if TYPE_CHECKING:
    from ..types import JSON_ro

from dataclasses import dataclass

from .types import RequestFiles, ExtendedRequestFiles


class Payload:
    pass

class Content(Payload):
    CONTENT_TYPE_HEADER_NAME: ClassVar[str] = 'Content-Type'
    MEDIA_TYPE_HINT: ClassVar[str] = 'application/octet-stream'

    def get_media_type(self) -> str:
        return self.MEDIA_TYPE_HINT

@dataclass(repr=False, eq=False)
class Bytes(Content):
    MEDIA_TYPE_HINT: ClassVar[str] = 'application/octet-stream'
    data: bytes

@dataclass(repr=False, eq=False)
class Text(Content):
    MEDIA_TYPE_HINT: ClassVar[str] = 'text/plain'
    text: str

@dataclass(repr=False, eq=False)
class URLEncodedFormData(Content):
    MEDIA_TYPE_HINT: ClassVar[str] = 'application/x-www-form-urlencoded'
    data: Mapping[str, str]

@dataclass(repr=False, eq=False)
class JSON(Content):
    MEDIA_TYPE_HINT: ClassVar[str] = 'application/json'
    json: JSON_ro

@dataclass(repr=False, eq=False)
class MultipartFormData(Content):
    @dataclass(repr=False, eq=False)
    class Field:
        name: str

    @dataclass(repr=False, eq=False)
    class TextField(Field):
        text: str

    @dataclass(repr=False, eq=False)
    class FileField(Field):
        file: IO[bytes]
        filename: Optional[str] = None
        content_type: Optional[str] = 'application/octet-stream'

    MEDIA_TYPE_HINT: ClassVar[str] = 'multipart/form-data'
    parts: Sequence[Field]



def make_multipart_payload_from_extended_request_files(xfiles: ExtendedRequestFiles) -> MultipartFormData:
    parts: list[MultipartFormData.Field] = []
    for key, value in xfiles.items():
        if isinstance(value, str):
            parts.append(MultipartFormData.TextField(key, value))

        elif isinstance(value, tuple):
            length = len(value)
            if length == 2:
                file, filename = cast("tuple[IO[bytes], str]", value)
                parts.append(MultipartFormData.FileField(key, file, filename))
            elif length == 3:
                file, filename, content_type = cast("tuple[IO[bytes], str, str]", value)
                parts.append(MultipartFormData.FileField(key, file, filename, content_type))
            else:
                raise TypeError

        else:
            parts.append(MultipartFormData.FileField(key, value))

    return MultipartFormData(parts)

def make_multipart(xfiles: ExtendedRequestFiles) -> MultipartFormData:
    return make_multipart_payload_from_extended_request_files(xfiles)


def make_multipart_payload_from_request_files(files: RequestFiles) -> MultipartFormData:
    parts: list[MultipartFormData.Field] = []
    for key, value in files.items():
        field: MultipartFormData.Field
        if isinstance(value, str):
            field = MultipartFormData.TextField(key, value)
        else:
            field = MultipartFormData.FileField(key, value)
        parts.append(field)
    return MultipartFormData(parts)

def make_payload(
    data: Optional[Union[Mapping[str, str], bytes]] = None,
    json: JSON_ro = None,
    files: Optional[RequestFiles] = None,
) -> Optional[Payload]:
    if files is not None:
        if json is not None:
            raise TypeError("mutually exclusive parameters: `files`, `json`")
        if data is not None:
            if isinstance(data, bytes):
                raise TypeError("`data` cannot be bytes when `files` is used")
            files = {**data, **files}
        return make_multipart_payload_from_request_files(files)

    if data is not None:
        if json is not None:
            raise TypeError("mutually exclusive parameters: `data`, `json`")
        if isinstance(data, bytes):
            return Bytes(data)
        return URLEncodedFormData(data)

    if json is not None:
        return JSON(json)

    return None




from __future__ import annotations
from typing import TYPE_CHECKING, TypeVar
if TYPE_CHECKING:
    from ..core.http_client_ASYNC import HTTPClient

from .core.http_client_ASYNC import build_reddit_http_client


class Client:
    _TSelf = TypeVar('_TSelf', bound='Client')

    @classmethod
    def from_http(cls: type[_TSelf], http: HTTPClient) -> _TSelf:
        self = cls.__new__(cls)
        self._init(http)
        return self

    def __init__(self) -> None:
        http = build_reddit_http_client()
        self._init(http)

    def _init(self, http: HTTPClient) -> None:
        self.http: HTTPClient = http
        ("")

        from .siteprocs.ASYNC import Procedures
        self.p: Procedures = Procedures(self.http)




from __future__ import annotations
from typing import Any, Optional

import sys
import urllib.request
import urllib.parse
import urllib.error
import socket
import json

from ... import exceptions
from ... import payload
from ...send_params import SendParams
from ...exchange import Exchange
from ...request import Request
from ...response import UResponse
from ..reg_SYNC import register
from ..connector_SYNC import Connector
from ...util.merge_query_params import merge_query_params


def _get_effective_timeout(v: float) -> float:
    if 0 > v != -1:
        raise ValueError(f"invalid `timeout` value: {v}")
    return v

def _get_effective_follow_redirects(v: Optional[bool]) -> bool:
    if v is None:
        raise ValueError(f"invalid `follow_redirects` value: {v}")
    return v


class PythonUrllibConnector(Connector):
    def _send(self, p: SendParams) -> Exchange:
        r = p.requisition

        etv = _get_effective_timeout(p.timeout)
        t = None if etv == -1 else etv

        follow_redirects = _get_effective_follow_redirects(p.follow_redirects)
        if not follow_redirects:
            raise RuntimeError('Python urllib does not support `follow_redirects=False`')

        url = merge_query_params(r.url, r.params)

        headers: dict[str, str] = dict(r.headers)
        data: Any = None

        pld = r.payload
        if pld is None:
            pass

        elif isinstance(pld, payload.Bytes):
            headers['Content-Type'] = pld.get_media_type()
            data = pld.data

        elif isinstance(pld, payload.Text):
            headers['Content-Type'] = pld.get_media_type()
            data = pld.text.encode()

        elif isinstance(pld, payload.JSON):
            headers['Content-Type'] = pld.get_media_type()
            data = json.dumps(pld.json).encode()

        elif isinstance(pld, payload.URLEncodedFormData):
            data = urllib.parse.urlencode(pld.data).encode()

        elif isinstance(pld, payload.MultipartFormData):
            raise Exception('multipart payload not supported by python urllib')

        else:
            raise Exception(f"unsupported payload type: {pld.__class__.__name__!r}")

        req = urllib.request.Request(
            method=r.verb,
            url=url,
            data=data,
            headers=headers,
        )
        try:
            with urllib.request.urlopen(req, timeout=t) as resp:
                content = resp.read()
        except socket.timeout as cause:
            raise exceptions.TimeoutException from cause
        except urllib.error.URLError as cause:
            if str(cause.reason) == 'timed out':
                raise exceptions.TimeoutException from cause
            raise
        except Exception as cause:
            raise exceptions.TransportError from cause

        x_requ = Request(
            verb=r.verb,
            url=url,
            headers={},
            data=b'',
        )
        x_resp = UResponse(
            status=resp.status,
            headers=resp.headers,
            data=content,
            underlying_object=resp,
        )
        return Exchange(
            requisition=r,
            request=x_requ,
            response=x_resp,
            history=(),
        )

    def _close(self) -> None:
        pass


def new_connector() -> PythonUrllibConnector:
    return PythonUrllibConnector()


name: str = 'python-urllib'
version: str = '%d.%d' % sys.version_info[:2]
register(
    adaptor_module_name=__name__,
    name=name,
    version=version,
    new_connector=new_connector,
)



"""WebSocket client."""

from . import const  # noqa: F401
from . import exceptions  # noqa: F401
from . import events  # noqa: F401
from . import utils  # noqa: F401
from . import transport  # noqa: F401




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Mapping
if TYPE_CHECKING:
    from ..client_SYNC import Client

from ..models.wiki import WikiPageRevisionAuthorUser as BaseWikiPageRevisionAuthorUser
from ..models.wiki_SYNC import WikiPageRevisionAuthorUser, WikiPage, WikiPageRevision, WikiPageSettings
from .wiki import (
    load_wiki_page as load_base_wiki_page,
    load_wiki_page_revision as load_base_wiki_page_revision,
    load_wiki_page_settings as load_base_wiki_page_settings,
)


def load_wiki_page_revision_author_user(d: Mapping[str, Any], client: Client) -> WikiPageRevisionAuthorUser:
    return WikiPageRevisionAuthorUser(d, client)


def promote_base_wiki_page_revision_author_user(base: BaseWikiPageRevisionAuthorUser, client: Client) -> WikiPageRevisionAuthorUser:
    return WikiPageRevisionAuthorUser(base.d, client)


def load_wiki_page(d: Mapping[str, Any], client: Client) -> WikiPage:
    up = load_base_wiki_page(d)
    return WikiPage(
        d=up.d,
        body=up.body,
        body_html=up.body_html,
        can_revise=up.can_revise,
        revision_uuid=up.revision_uuid,
        revision_unixtime=up.revision_unixtime,
        revision_author=promote_base_wiki_page_revision_author_user(up.revision_author, client),
        revision_message=up.revision_message,
    )

def load_wiki_page_revision(d: Mapping[str, Any], client: Client) -> WikiPageRevision:
    up = load_base_wiki_page_revision(d)
    return WikiPageRevision(
        d=up.d,
        uuid=up.uuid,
        unixtime=up.unixtime,
        author=promote_base_wiki_page_revision_author_user(up.author, client),
        message=up.message,
        hidden=up.hidden,
    )

def load_wiki_page_settings(d: Mapping[str, Any], client: Client) -> WikiPageSettings:
    up = load_base_wiki_page_settings(d)
    return WikiPageSettings(
        permlevel=up.permlevel,
        editors=[promote_base_wiki_page_revision_author_user(o, client) for o in up.editors],
        indexed=up.indexed,
    )




from typing import Any, Mapping

from datetime import datetime
from dataclasses import dataclass

from .datamemento import DatamementoDataclassesMixin

@dataclass(repr=False, eq=False)
class ModerationActionLogEntry(DatamementoDataclassesMixin):
    #_: KW_ONLY
    d: Mapping[str, Any]
    uuid: str
    ("""
        Mod action unique identifier.
        """)
    unixtime: int
    ("""
        Unix timestamp of when the action was performed.
        """)
    datetime: datetime
    ("""
        Datetime object of when the action was performed.
        """)
    action: str
    ("""
        Name of the moderation action.
        """)
    agent_id: int
    ("""
        ID of the moderator who initiated the action.
        """)
    agent: str
    ("""
        Name of the moderator who initiated the action.
        """)




from __future__ import annotations
from typing import Any, Mapping

from ..models.moderated_subreddit import ModeratedSubreddit

def load_moderated_subreddit(d: Mapping[str, Any]) -> ModeratedSubreddit:
    return ModeratedSubreddit(d)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.submission_collection_ASYNC import SubmissionCollectionInfo, SubmissionCollection

from ...model_loaders.submission_collection_ASYNC import load_submission_collection_info, load_submission_collection
from ...util.base_conversion import to_base36


class CollectionProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client

    async def create(self,
        sr_id: int,
        title: str,
        description: Optional[str] = None,
        display_layout: Optional[str] = None,
    ) -> SubmissionCollection:
        """
        Create a collection.

        Returns the newly created collection.

        .. .PARAMETERS

        :param `int` sr_id:
            Subreddit ID.
        :param `str` title:
            A string no longer than 300 characters.
        :param `Optional[str]` description:
            A string no longer than 500 characters.
        :param `Optional[str]` display_layout:
            Either `TIMELINE` or `GALLERY`.

            Default is `TIMELINE`.

        .. .RETURNS

        :rtype: :class:`~.models.submission_collection_ASYNC.SubmissionCollection`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `SUBREDDIT_NOEXIST`:
                The specified subreddit (`sr_id`) does not exist.
            + `NO_TEXT`:
                The specified `title` was empty.
            + `TOO_LONG`:
               - The specified `title` was longer than 300 characters.
               - The specified `description` was longer than 500 characters.

            + `INVALID_OPTION`:
                The value specified for `display_layout` is not valid.

                The options are case-sensitive.
        """
        sr_id36 = to_base36(sr_id)
        params = {'sr_fullname': 't5_' + sr_id36, 'title': title}
        if description is not None:
            params['description'] = description
        if display_layout is not None:
            params['display_layout'] = display_layout
        data = await self._client.request('POST', '/api/v1/collections/create_collection', params=params)
        return load_submission_collection(data, self._client)

    async def add_post(self, uuid: str, subm_id: int) -> None:
        """
        Add a submission to a collection.

        Collections have a capacity of 100 posts. Attempting to add to a full
        collection will result in an `INVALID_COLLECTION_UPDATE` API error.

        .. .PARAMETERS

        :param `str` uuid:
        :param `int` subm_id:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_TEXT`:
                The specified `uuid` was empty.
            + `TOO_SHORT`:
                The specified `uuid` was under 36 characters.
            + `TOO_LONG`:
               - The specified `uuid` was over 36 characters.
               - The specified `description` was longer than 500 characters.

            + `INVALID_COLLECTION_UPDATE`:
               - The specified `uuid` does not exist.
               - The submission specified by `subm_id` already exists in
                 the collection.
               - The submission specified by `subm_id` does not match
                 the collection's subreddit.
               - The collection is full (it contains 100 posts).

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `404`:
                The submission specified by `subm_id` does not exist.
            + `500`:
                The value specified by `uuid` is not a valid UUID.
        """
        id36 = to_base36(subm_id)
        params = {'collection_id': uuid, 'link_fullname': 't3_' + id36}
        await self._client.request('POST', '/api/v1/collections/add_post_to_collection', params=params)

    async def remove_post(self, uuid: str, subm_id: int) -> None:
        """
        Remove a submission from a collection.

        .. .PARAMETERS

        :param `str` uuid:
        :param `int` subm_id:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_TEXT`:
                The specified `uuid` was empty.
            + `TOO_SHORT`:
                The specified `uuid` was under 36 characters.
            + `TOO_LONG`:
               - The specified `uuid` was over 36 characters.
               - The specified `description` was longer than 500 characters.

            + `INVALID_COLLECTION_UPDATE`:
               - The specified `uuid` does not exist.
               - The submission specified by `subm_id` does not exist in
                 the collection.

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `404`:
               The submission specified by `subm_id` does not exist.
            + `500`:
               The value specified by `uuid` is not a valid UUID.
        """
        id36 = to_base36(subm_id)
        params = {'collection_id': uuid, 'link_fullname': 't3_' + id36}
        await self._client.request('POST', '/api/v1/collections/remove_post_in_collection', params=params)

    async def get_full(self, uuid: str) -> Optional[SubmissionCollection]:
        """Get a collection, including its submissions.

        .. .PARAMETERS

        :param `str` uuid:

        .. .RETURNS

        :rtype: `Optional`\\[:class:`~.models.submission_collection_ASYNC.SubmissionCollection`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `NO_TEXT`:
                The specified `uuid` was empty.
            + `TOO_SHORT`:
                The specified `uuid` was under 36 characters.
            + `TOO_LONG`:
                The specified `uuid` was over 36 characters.
        """
        params = {'collection_id': uuid}
        root = await self._client.request('GET', '/api/v1/collections/collection', params=params)
        if len(root) < 3:
            return None
        return load_submission_collection(root, self._client)

    async def get_info(self, uuid: str) -> Optional[SubmissionCollectionInfo]:
        """Get a collection, excluding its submissions.

        .. .PARAMETERS

        :param `str` uuid:

        .. .RETURNS

        :rtype: `Optional`\\[:class:`~.models.submission_collection_ASYNC.SubmissionCollectionInfo`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `NO_TEXT`:
                The specified `uuid` was empty.
            + `TOO_SHORT`:
                The specified `uuid` was under 36 characters.
            + `TOO_LONG`:
                The specified `uuid` was over 36 characters.
        """
        params = {'collection_id': uuid, 'include_links': '0'}
        root = await self._client.request('GET', '/api/v1/collections/collection', params=params)
        if len(root) < 3:
            return None
        return load_submission_collection_info(root, self._client)

    async def get_subreddit_collections_info(self, sr_id: int) -> Sequence[SubmissionCollectionInfo]:
        """
        Get a list of collections' details from the subreddit.

        .. .PARAMETERS

        :param `int` id:
            Subreddit ID.

        .. .RETURNS

        :rtype: `Sequence`\\[:class:`~.models.submission_collection_ASYNC.SubmissionCollectionInfo`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `SUBREDDIT_NOEXIST`:
                The specified subreddit could not be found.
        """
        id36 = to_base36(sr_id)
        params = {'sr_fullname': 't5_' + id36}
        root = await self._client.request('GET', '/api/v1/collections/subreddit_collections', params=params)
        return [load_submission_collection_info(d, self._client) for d in root]

    async def delete(self, uuid: str) -> None:
        """Delete a collection.

        .. .PARAMETERS

        :param `str` uuid:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_TEXT`:
                The specified `uuid` was empty.
            + `TOO_SHORT`:
                The specified `uuid` was under 36 characters.
            + `TOO_LONG`:
                The specified `uuid` was over 36 characters.
            + `INVALID_COLLECTION_ID`:
                The specified does not exist.
        """
        params = {'collection_id': uuid}
        await self._client.request('POST', '/api/v1/collections/delete_collection', params=params)

    async def reorder(self, uuid: str, subm_ids: Sequence[int]) -> None:
        """
        Reorder posts in a collection.

        An API error is returned (`INVALID_COLLECTION_UPDATE`) if an ID in
        the given list is not found in the collection.

        If only a subset of the IDs in the collection are specified then those
        submissions will be moved to the top of the collection in the order specified.
        The rest are moved down, maintaining their order.

        .. .PARAMETERS

        :param `str` uuid:
        :param `Sequence[int]` subm_ids:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `INVALID_COLLECTION_UPDATE`:
                One of the IDs specified in the `subm_ids` list does not
                exist in the collection.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `404`:
               - The specified `uuid` was empty.
               - The specified `uuid` is invalid.
               - The specified `uuid` does not exist.
        """
        id36s = [to_base36(i) for i in subm_ids]
        full_id36s = ['t3_' + s for s in id36s]
        params = {'collection_id': uuid, 'link_ids': ','.join(full_id36s)}
        await self._client.request('POST', '/api/v1/collections/reorder_collection', params=params)

    async def set_title(self, uuid: str, title: str) -> None:
        """Update a collection's title.

        .. .PARAMETERS

        :param `str` uuid:
        :param `str` title:
            New title for the collection, up to 300 characters long.

            It should not be an empty string.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_TEXT`:
               - The specified `uuid` was empty.
               - The specified `title` was empty.

            + `TOO_SHORT`:
                The specified `uuid` was under 36 characters.
            + `TOO_LONG`:
               - The specified `uuid` was over 36 characters.
               - The specified `title` was over 300 characters.

            + `INVALID_COLLECTION_ID`:
                The specified `uuid` does not exist.
        """
        params = {'collection_id': uuid, 'title': title}
        await self._client.request('POST', '/api/v1/collections/update_collection_title', params=params)

    async def set_description(self, uuid: str, desc: str) -> None:
        """Update a collection's description.

        .. .PARAMETERS

        :param `str` uuid:
        :param `str` title:
            New description for the collection, up to 500 characters long.

            Can be an empty string.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_TEXT`:
               - The specified `uuid` was empty.
               - The specified `title` was empty.

            + `TOO_SHORT`:
                The specified `uuid` was under 36 characters.
            + `TOO_LONG`:
               - The specified `uuid` was over 36 characters.
               - The specified `description` was over 500 characters.

            + `INVALID_COLLECTION_ID`:
                The specified `uuid` does not exist.
        """
        params = {'collection_id': uuid, 'description': desc}
        await self._client.request('POST', '/api/v1/collections/update_collection_description', params=params)

    async def set_display_layout(self, uuid: str, layout: Optional[str]) -> None:
        """Update a collection's display layout.

        .. .PARAMETERS

        :param `str` uuid:
        :param `Optional[str]` layout:
            Either `TIMELINE` or `GALLERY`.

            Case-sensitive.

            If not specified or an empty string, the `display_layout` field on the collection
            object will be set to null, which is treated the same as `TIMELINE`.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_TEXT`:
                The specified `uuid` was empty.
            + `TOO_SHORT`:
                The specified `uuid` was under 36 characters.
            + `TOO_LONG`:
                The specified `uuid` was over 36 characters.
            + `INVALID_COLLECTION_ID`:
                The specified `uuid` does not exist.
            + `INVALID_OPTION`:
                The value specified for `display_layout` is not valid.

                The options are case-sensitive.
        """
        params = {'collection_id': uuid}
        if layout is not None:
            params['display_layout'] = layout
        await self._client.request('POST', '/api/v1/collections/update_collection_display_layout', params=params)

    async def follow(self, uuid: str) -> None:
        """Follow a collection.

        This actually has a similar effect to
        :meth:`~.siteprocs.submission.SYNC.SubmissionProcedures.follow_event`.
        When you follow a collection, all submissions in the collection will be followed,
        *even if there is no event information* on any of the posts.

        When a new submission is added to a followed collection, the submission is
        automatically followed. If any of the individual submissions are unfollowed via
        :meth:`~.siteprocs.submission.SYNC.SubmissionProcedures.unfollow_event`
        then the whole collection and all of its submissions are unfollowed,
        having the same effect as :meth:`.unfollow`.

        To tell if a collection is followed, check if one of its submissions is::

           coll = await client.p.collection.get_full('84359211-be58-4c98-87cd-26bc10c59fb3')
           if coll is None:
               raise Exception
           followed = any(subm.me.is_following_event for subm in coll)
           print(followed)

        .. .PARAMETERS

        :param `str` uuid:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_TEXT`:
                The specified `uuid` was empty.
            + `TOO_SHORT`:
                The specified `uuid` was under 36 characters.
            + `TOO_LONG`:
                The specified `uuid` was over 36 characters.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `500`:
                The specified `uuid` does not exist.
        """
        params = {'follow': '1', 'collection_id': uuid}
        await self._client.request('POST', '/api/v1/collections/follow_collection', params=params)

    async def unfollow(self, uuid: str) -> None:
        """Unfollow a collection.

        .. .PARAMETERS

        :param `str` uuid:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :(raises):
            (Same as :meth:`.follow`.)
        """
        params = {'follow': '0', 'collection_id': uuid}
        await self._client.request('POST', '/api/v1/collections/follow_collection', params=params)




from __future__ import annotations
from typing import Optional

import asyncio

import aiohttp  # type: ignore[import]

from ... import exceptions
from ... import payload
from ...send_params import SendParams
from ...exchange import Exchange
from ...request import Request
from ...response import UResponse
from ..reg_ASYNC import register
from ..connector_ASYNC import Connector


def _get_effective_timeout(v: float) -> float:
    if 0 > v != -1:
        raise ValueError(f"invalid `timeout` value: {v}")
    return v

def _get_effective_follow_redirects(v: Optional[bool]) -> bool:
    if v is None:
        raise ValueError(f"invalid `follow_redirects` value: {v}")
    return v


class AiohttpConnector(Connector):
    def __init__(self,
        session: aiohttp.ClientSession,
    ) -> None:
        super().__init__()
        self.session: aiohttp.ClientSession = session
        ("")

    async def _send(self, p: SendParams) -> Exchange:
        r = p.requisition

        etv = _get_effective_timeout(p.timeout)

        client_timeout = aiohttp.ClientTimeout(
            total=etv,
            connect=etv,
            sock_connect=etv,
            sock_read=etv,
        )
        if etv == -1:
            client_timeout = aiohttp.ClientTimeout(
                total=None,
                connect=None,
                sock_connect=None,
                sock_read=None,
            )

        follow_redirects = _get_effective_follow_redirects(p.follow_redirects)

        headers: dict[str, str] = dict(r.headers)
        data: object = None
        json: object = None

        pld = r.payload
        if pld is None:
            pass

        elif isinstance(pld, payload.Bytes):
            headers['Content-Type'] = pld.get_media_type()
            data = pld.data

        elif isinstance(pld, payload.Text):
            headers['Content-Type'] = pld.get_media_type()
            data = pld.text.encode()

        elif isinstance(pld, payload.JSON):
            json = pld.json

        elif isinstance(pld, payload.URLEncodedFormData):
            data = dict(pld.data)

        elif isinstance(pld, payload.MultipartFormData):
            data = aiohttp.FormData()
            for pt in pld.parts:
                if isinstance(pt, payload.MultipartFormData.TextField):
                    data.add_field(pt.name, pt.text)
                elif isinstance(pt, payload.MultipartFormData.FileField):
                    data.add_field(pt.name, pt.file, filename=pt.filename, content_type=pt.content_type)
                else:
                    raise ValueError('unexpected multipart field type: ' + repr(pt))

        else:
            raise Exception(f"unsupported payload type: {pld.__class__.__name__!r}")

        try:
            async with self.session.request(
                method=r.verb,
                url=r.url,
                params=r.params,
                headers=headers,
                timeout=client_timeout,
                allow_redirects=follow_redirects,
                data=data,
                json=json,
            ) as resp:
                requ = resp.request_info
                x_requ = Request(
                    verb=requ.method,
                    url=str(requ.real_url),
                    headers=requ.headers,
                    data=b'',
                )
                x_resp = UResponse(
                    status=resp.status,
                    headers=resp.headers,
                    data=(await resp.content.read()),
                    underlying_object=resp,
                )
                history = [
                    UResponse(
                        status=resp1.status,
                        headers=resp1.headers,
                        data=(await resp1.content.read()),
                        underlying_object=resp1,
                    )
                    for resp1 in resp.history
                ]
        except asyncio.TimeoutError as cause:
            raise exceptions.TimeoutException from cause
        except Exception as cause:
            raise exceptions.TransportError from cause

        return Exchange(
            requisition=r,
            request=x_requ,
            response=x_resp,
            history=history,
        )

    async def _close(self) -> None:
        await self.session.close()


def new_connector() -> AiohttpConnector:
    return AiohttpConnector(aiohttp.ClientSession())


name: str = aiohttp.__name__
version: str = aiohttp.__version__
register(
    adaptor_module_name=__name__,
    name=name,
    version=version,
    new_connector=new_connector,
)



"""HTTP client."""

from . import send_params  # noqa: F401
from .send_params import SendParams as SendParams  # noqa: F401
from . import exchange  # noqa: F401
from .exchange import Exchange as Exchange  # noqa: F401
from . import requisition  # noqa: F401
from .requisition import Requisition as Requisition, make_requisition as make_requisition  # noqa: F401
from . import request  # noqa: F401
from .request import Request as Request  # noqa: F401
from . import response  # noqa: F401
from .response import Response as Response  # noqa: F401
from . import exceptions  # noqa: F401
from . import payload  # noqa: F401
from . import misc  # noqa: F401
from . import transport  # noqa: F401
from . import util  # noqa: F401




from __future__ import annotations
from typing import TYPE_CHECKING, TypeVar, Optional, Iterable, Callable, Any, Mapping
if TYPE_CHECKING:
    from ....client_SYNC import Client

from ..listing.listing_paginator import ListingPaginator
from ..listing.mixins.subreddit_detail_SYNC import SubredditDetail
from ..listing.submission_and_comment_listing_paginator import SubmissionAndLooseCommentListingPaginator
from ..listing.submission_listing_paginator import SubmissionListingPaginator
from ..listing.comment_listing_paginator import LooseCommentListingPaginator
from ....models.submission_SYNC import Submission
from ....models.comment_SYNC import LooseComment

T = TypeVar('T')

class OnlyFilter(ListingPaginator[T]):
    def __init__(self,
        client: Client,
        url: str,
        *,
        limit: Optional[int] = 100,
        params: Optional[Mapping[str, str]] = None,
        cursor_extractor: Callable[[Any], str] = lambda x: x['data']['name'],
    ) -> None:
        super().__init__(client, url, limit=limit, params=params, cursor_extractor=cursor_extractor)
        self._only = ''

    def _generate_params(self) -> Iterable[tuple[str, str]]:
        yield from super()._generate_params()
        if self._only:
            yield ('only', self._only)

class SubmissionOnlyFilter(OnlyFilter[Submission]):
    def __init__(self,
        client: Client,
        url: str,
        *,
        limit: Optional[int] = 100,
        params: Optional[Mapping[str, str]] = None,
        cursor_extractor: Callable[[Any], str] = lambda x: x['data']['name'],
    ) -> None:
        super().__init__(client, url, limit=limit, params=params, cursor_extractor=cursor_extractor)
        self._only = 'links'

class CommentOnlyFilter(OnlyFilter[LooseComment]):
    def __init__(self,
        client: Client,
        url: str,
        *,
        limit: Optional[int] = 100,
        params: Optional[Mapping[str, str]] = None,
        cursor_extractor: Callable[[Any], str] = lambda x: x['data']['name'],
    ) -> None:
        super().__init__(client, url, limit=limit, params=params, cursor_extractor=cursor_extractor)
        self._only = 'comments'


class ModQueueListingPaginator(
    SubredditDetail[object],
    SubmissionAndLooseCommentListingPaginator,
): pass
class ModQueueSubmissionListingPaginator(
    SubredditDetail[Submission],
    SubmissionOnlyFilter,
    SubmissionListingPaginator,
): pass
class ModQueueCommentListingPaginator(
    SubredditDetail[LooseComment],
    CommentOnlyFilter,
    LooseCommentListingPaginator,
): pass

class ReportsListingPaginator(
    SubredditDetail[object],
    SubmissionAndLooseCommentListingPaginator,
): pass
class ReportsSubmissionListingPaginator(
    SubredditDetail[Submission],
    SubmissionOnlyFilter,
    SubmissionListingPaginator,
): pass
class ReportsCommentListingPaginator(
    SubredditDetail[LooseComment],
    CommentOnlyFilter,
    LooseCommentListingPaginator,
): pass

class SpamListingPaginator(
    SubredditDetail[object],
    SubmissionAndLooseCommentListingPaginator,
): pass
class SpamSubmissionListingPaginator(
    SubredditDetail[Submission],
    SubmissionOnlyFilter,
    SubmissionListingPaginator,
): pass
class SpamCommentListingPaginator(
    SubredditDetail[LooseComment],
    CommentOnlyFilter,
    LooseCommentListingPaginator,
): pass

class EditedListingPaginator(
    SubredditDetail[object],
    SubmissionAndLooseCommentListingPaginator,
): pass
class EditedSubmissionListingPaginator(
    SubredditDetail[Submission],
    SubmissionOnlyFilter,
    SubmissionListingPaginator,
): pass
class EditedCommentListingPaginator(
    SubredditDetail[LooseComment],
    CommentOnlyFilter,
    LooseCommentListingPaginator,
): pass

class UnmoderatedSubmissionListingPaginator(
    SubredditDetail[Submission],
    SubmissionListingPaginator,
): pass




from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Collection, Optional, Sequence
if TYPE_CHECKING:
    from ..client_SYNC import Client

from ...iterators.call_chunk_chaining_iterator import CallChunkChainingIterator
from ...pagination.paginator_chaining_iterator import PaginatorChainingIterator
from ...util.base_conversion import to_base36
from ...iterators.call_chunk import CallChunk
from ...iterators.chunking import chunked
from ..paginators.document_paginator import DocumentPaginator
from ..models.document import Document


class Procedures:
    def __init__(self, client: Client) -> None:
        self._client = client

    def bulk_fetch_submissions(self, ids: Iterable[int], *, fields: Optional[Collection[str]] = None) -> CallChunkChainingIterator[Document]:
        """Bulk fetch submission documents by ID.

        .. .PARAMETERS

        :param `Iterable[int]` ids:
            Submission IDs.
        :param `Optional[Collection[str]]` fields:
            Return only specific fields.

        .. .RETURNS

        :rtype: `~.iterators.call_chunk_chaining_iterator.CallChunkChainingIterator`\\[:class:`~.pushshift.models.document.Document`]
        """
        def mass_fetch_submissions(ids: Sequence[int]) -> Sequence[Document]:
            def g() -> Iterable[tuple[str, str]]:
                yield ('limit', '1000')
                yield ('ids', ','.join(map(to_base36, ids)))
                if fields: yield ('fields', ','.join(fields))
            root = self._client.request('GET', '/reddit/submission/search', params=dict(g()))
            return [Document(d) for d in root['data']]

        return CallChunkChainingIterator(CallChunk(mass_fetch_submissions, chunk) for chunk in chunked(ids, 1000))

    def count_search_submissions(self,
        *,
        query: Optional[str] = None,
        query_title: Optional[str] = None,
        query_body: Optional[str] = None,
        subreddit: Optional[str] = None,
        author: Optional[str] = None,
        time_range: Optional[tuple[Optional[int], Optional[int]]] = None,
        since: Optional[int] = None,
        until: Optional[int] = None,
    ) -> int:
        """Return the number of expected entries for a submission search.

        .. .PARAMETERS

        :(parameters): Similar to :meth:`.page_search_submissions`.

        .. .RETURNS

        :rtype: `int`
        """
        if time_range is not None:
            if (since, until) != (None, None):
                raise TypeError("`time_range` cannot be used with `since` or `until`")
            since, until = time_range

        def g() -> Iterable[tuple[str, str]]:
            yield ('track_total_hits', '1')
            yield ('limit', '0')
            if since is not None: yield ('since', str(since))
            if until is not None: yield ('until', str(until))
            if query is not None: yield ('q', query)
            if query_title is not None: yield ('title', query_title)
            if query_body is not None: yield ('selftext', query_body)
            if subreddit is not None: yield ('subreddit', subreddit)
            if author is not None: yield ('author', author)

        root = self._client.request('GET', '/reddit/submission/search', params=dict(g()))
        return root['metadata']['es']['hits']['total']['value']

    def page_search_submissions(self,
        *,
        query: Optional[str] = None,
        query_title: Optional[str] = None,
        query_body: Optional[str] = None,
        subreddit: Optional[str] = None,
        author: Optional[str] = None,
        time_range: Optional[tuple[Optional[int], Optional[int]]] = None,
        since: Optional[int] = None,
        until: Optional[int] = None,
        fields: Optional[Collection[str]] = None,
        descending: bool = False,
    ) -> DocumentPaginator:
        """Return a paginator for a submission search.

        .. .PARAMETERS

        :param `Optional[str]` query:
            Filter by title and body.
        :param `Optional[str]` query_title:
            Filter by title.
        :param `Optional[str]` query_body:
            Filter by body.
        :param `Optional[str]` subreddit:
            Filter by specific subreddits.

            A comma-separated list is accepted.
        :param `Optional[str]` author:
            Filter by specific authors.

            A comma-separated list is accepted.
        :param `Optional[tuple[Optional[int], Optional[int]]]` time_range:
            Filter by time.

            Takes a time range tuple of `(since, until)`.
            The `since` value is inclusive.
            The `until` value is exclusive.

            The `until` value should always be greater than `since`.

            This parameter cannot be used with `since` or `until`.
        :param `Optional[int]` since:
        :param `Optional[int]` until:
        :param `Optional[Collection[str]]` fields:
            Return only specific fields.

            Values `id` and `created_utc` MUST be included otherwise
            the pagination logic will error out.
        :param `Optional[str]` descending:
            Return results in descending order.

            Note, this only controls the order of the entries according to the
            `created_utc` field and not the `id` field. The order of `id`\\ s is
            never guarented. This means that if you collect, reverse and compare
            the order of entries' `id` fields with a different `descending` value,
            they may not be equal.

        .. .RETURNS

        :rtype: :class:`~.pushshift.paginators.document_paginator.DocumentPaginator`
        """
        if time_range is not None:
            if (since, until) != (None, None):
                raise TypeError("`time_range` cannot be used with `since` or `until`")
            since, until = time_range

        def g() -> Iterable[tuple[str, str]]:
            if query is not None: yield ('q', query)
            if query_title is not None: yield ('title', query_title)
            if query_body is not None: yield ('selftext', query_body)
            if subreddit is not None: yield ('subreddit', subreddit)
            if author is not None: yield ('author', author)

        return DocumentPaginator(
            client=self._client,
            url='/reddit/submission/search',
            search_params=dict(g()),
            since=since,
            until=until,
            fields=fields,
            descending=descending,
        )

    def pull_search_submissions(self,
        *,
        amount: Optional[int] = None,
        query: Optional[str] = None,
        query_title: Optional[str] = None,
        query_body: Optional[str] = None,
        subreddit: Optional[str] = None,
        author: Optional[str] = None,
        time_range: Optional[tuple[Optional[int], Optional[int]]] = None,
        since: Optional[int] = None,
        until: Optional[int] = None,
        fields: Optional[Collection[str]] = None,
        descending: bool = False,
    ) -> PaginatorChainingIterator[Document]:
        """Pull entries from a submission search.

        Using a paginator object directly is generally recommended.
        I.e., :meth:`.page_search_submissions` is preferred over this method.
        """
        paginator = self.page_search_submissions(
            query=query,
            query_title=query_title,
            query_body=query_body,
            subreddit=subreddit,
            author=author,
            time_range=time_range,
            since=since,
            until=until,
            fields=fields,
            descending=descending,
        )
        return PaginatorChainingIterator(paginator, amount=amount)


    def bulk_fetch_comments(self, ids: Iterable[int], *, fields: Optional[Collection[str]] = None) -> CallChunkChainingIterator[Document]:
        """Bulk fetch comment documents by ID.

        Behaves similarly to :meth:`.bulk_fetch_submissions`.
        """
        def mass_fetch_comments(ids: Sequence[int]) -> Sequence[Document]:
            def g() -> Iterable[tuple[str, str]]:
                yield ('limit', '1000')
                yield ('ids', ','.join(map(to_base36, ids)))
                if fields: yield ('fields', ','.join(fields))
            root = self._client.request('GET', '/reddit/comment/search', params=dict(g()))
            return [Document(d) for d in root['data']]

        return CallChunkChainingIterator(CallChunk(mass_fetch_comments, chunk) for chunk in chunked(ids, 1000))

    def count_search_comments(self,
        *,
        query: Optional[str] = None,
        subreddit: Optional[str] = None,
        author: Optional[str] = None,
        time_range: Optional[tuple[Optional[int], Optional[int]]] = None,
        since: Optional[int] = None,
        until: Optional[int] = None,
    ) -> int:
        """Return the number of expected entries for a submission search.

        Behaves similarly to :meth:`.count_search_submissions`.
        """
        if time_range is not None:
            if (since, until) != (None, None):
                raise TypeError("`time_range` cannot be used with `since` or `until`")
            since, until = time_range

        def g() -> Iterable[tuple[str, str]]:
            yield ('track_total_hits', '1')
            yield ('limit', '0')
            if since is not None: yield ('since', str(since))
            if until is not None: yield ('until', str(until))
            if query is not None: yield ('q', query)
            if subreddit is not None: yield ('subreddit', subreddit)
            if author is not None: yield ('author', author)

        root = self._client.request('GET', '/reddit/comment/search', params=dict(g()))
        return root['metadata']['es']['hits']['total']['value']

    def page_search_comments(self,
        *,
        query: Optional[str] = None,
        subreddit: Optional[str] = None,
        author: Optional[str] = None,
        time_range: Optional[tuple[Optional[int], Optional[int]]] = None,
        since: Optional[int] = None,
        until: Optional[int] = None,
        fields: Optional[Collection[str]] = None,
        descending: bool = False,
    ) -> DocumentPaginator:
        """Return a paginator for a comment search.

        Behaves similarly to :meth:`.page_search_submissions`.
        """
        if time_range is not None:
            if (since, until) != (None, None):
                raise TypeError("`time_range` cannot be used with `since` or `until`")
            since, until = time_range

        def g() -> Iterable[tuple[str, str]]:
            if query is not None: yield ('q', query)
            if subreddit is not None: yield ('subreddit', subreddit)
            if author is not None: yield ('author', author)

        return DocumentPaginator(
            client=self._client,
            url='/reddit/comment/search',
            search_params=dict(g()),
            since=since,
            until=until,
            fields=fields,
            descending=descending,
        )

    def pull_search_comments(self,
        *,
        amount: Optional[int] = None,
        query: Optional[str] = None,
        subreddit: Optional[str] = None,
        author: Optional[str] = None,
        time_range: Optional[tuple[Optional[int], Optional[int]]] = None,
        since: Optional[int] = None,
        until: Optional[int] = None,
        fields: Optional[Collection[str]] = None,
        descending: bool = False,
    ) -> PaginatorChainingIterator[Document]:
        """Pull entries from a comment search.

        Behaves similarly to :meth:`.pull_search_submissions`.
        """
        paginator = self.page_search_comments(
            query=query,
            subreddit=subreddit,
            author=author,
            time_range=time_range,
            since=since,
            until=until,
            fields=fields,
            descending=descending,
        )
        return PaginatorChainingIterator(paginator, amount=amount)




from __future__ import annotations
from typing import TYPE_CHECKING, TypeVar, Optional, Sequence
if TYPE_CHECKING:
    from ....client_SYNC import Client

from operator import itemgetter

from ..listing.listing_paginator import ListingPaginator
from ....model_loaders.user_relationship import load_user_relationship, load_banned_subreddit_user_relation
from ....models.user_relationship import UserRelationship, BannedSubredditUserRelationship

T = TypeVar('T')

class LegacyModerationUsersPaginator(ListingPaginator[T]):
    def __init__(self,
        client: Client,
        url: str,
        *,
        limit: Optional[int] = 100,
    ) -> None:
        super().__init__(client, url, limit=limit, cursor_extractor=itemgetter('rel_id'))


class UserRelationshipListingPaginator(LegacyModerationUsersPaginator[UserRelationship]):
    def fetch(self) -> Sequence[UserRelationship]:
        data = self._fetch_data()
        return [load_user_relationship(d) for d in data['children']]

class BannedSubredditUserRelationshipListingPaginator(LegacyModerationUsersPaginator[BannedSubredditUserRelationship]):
    def fetch(self) -> Sequence[BannedSubredditUserRelationship]:
        data = self._fetch_data()
        return [load_banned_subreddit_user_relation(d) for d in data['children']]




from __future__ import annotations
from typing import Optional, Sequence

from .praw_config import get_praw_config
from ..client_ASYNC import Client

def client_from_praw_config(section_name: str, *, filepath: Optional[str] = None) -> Client:
    """See :meth:`client_ASYNC.Client.from_praw_config`."""
    config = get_praw_config(filepath)
    section_name = section_name or config.default_section
    try:
        section = config[section_name]
    except KeyError:
        empty = not any(s.values() for s in config.values())
        msg = f"No section named {section_name!r} in{' empty' if empty else ''} praw.ini config."
        class StrReprStr(str):
            def __repr__(self) -> str:
                return str(self)
        raise KeyError(StrReprStr(msg)) from None

    get = section.get
    grant_creds: Sequence[str] = ()
    if refresh_token := get('refresh_token'):
        grant_creds = (refresh_token,)
    elif (username := get('username')) and (password := get('password')):
        grant_creds = (username, password)
    client = Client(
        section['client_id'],
        section['client_secret'],
        *grant_creds,
    )
    if x := get('user_agent'):
        client.set_user_agent(x)
    return client




from __future__ import annotations
from typing import Any, Mapping, Optional

from functools import cached_property
from datetime import datetime, timezone

class UserRelationship:
    @cached_property
    def added_at(self) -> datetime:
        """Datetime object of when the relationship was created."""
        return datetime.fromtimestamp(self.added_ut, timezone.utc)

    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        ("")
        full_id36: str = d['id']
        _, _, id36 = full_id36.partition('_')
        self.id36: str = id36
        ("""
            User ID as a base 36 number.
            """)
        self.id: int = int(id36, 36)
        ("""
            User ID.
            """)
        self.name: str = d['name']
        ("""
            Username.
            """)
        #self.rel_id: str = d['rel_id']
        self.added_ut: int = int(d['date'])
        ("""
            UNIX timestamp of when the relationship was created.
            """)

class FriendRelationship(UserRelationship):
    def __init__(self, d: Mapping[str, Any]) -> None:
        super().__init__(d)
        # Need Reddit Premium to test this
        # self.note: str = d['note'] or ''

class BannedSubredditUserRelationship(UserRelationship):
    def __init__(self, d: Mapping[str, Any]) -> None:
        super().__init__(d)
        self.days_remaining: Optional[int] = d['days_left']
        ("""
            Number of days until the ban is lifted.

            Value `None` if the ban is permanent.
            """)
        self.reason: str = d['note']
        ("""
            Ban reason.
            """)




from __future__ import annotations

from ...core.http_client_ASYNC import HTTPClient
from ...http.util.json_loading import load_json_from_response
from .login.ASYNC import LoginProcedures


class Procedures:
    def __init__(self, http: HTTPClient) -> None:
        self._http = http
        self.login: LoginProcedures = LoginProcedures(http)
        ("")

    async def ping(self) -> None:
        await self._http.request('GET', 'https://oauth.reddit.com/api/v1/scopes', params={'scopes': 'read'})

    async def get_sendbird_access_token(self) -> tuple[str, int]:
        resp = await self._http.request('GET', 'https://sendbird.reddit.com/api/v1/sendbird/me')
        resp.ensure_successful_status()
        d = load_json_from_response(resp)
        access_token = d['sb_access_token']
        expires_utms = d['valid_until']
        expires_ts = expires_utms // 1000
        return (access_token, expires_ts)




from __future__ import annotations
from typing import TYPE_CHECKING, IO, Optional
if TYPE_CHECKING:
    from ...client_ASYNC import Client

import os.path as op
from functools import cached_property

from ...models.flair_emoji import FlairEmojiUploadLease, SubredditFlairEmojis
from ...model_loaders.flair_emoji import load_flair_emoji_upload_lease, load_flair_emoji
from ...http.util.guess_filename_mimetype import guess_filename_mimetype

class FlairEmojiProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client

    async def retrieve(self, sr: str) -> SubredditFlairEmojis:
        """Get a list of all flair emojis in a subreddit.

        .. .PARAMETERS

        :param `str` sr:

        .. .RETURNS

        :returns: A mapping from flair emoji name to
            :class:`~.models.flair_emoji.FlairEmoji`.
        :rtype: :class:`~.models.flair_emoji.SubredditFlairEmojis`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                The specified subreddit cannot be accessed.
            + `500`:
                The specified subreddit does not exist.
        """
        root = await self._client.request('GET', f'/api/v1/{sr}/emojis/all')
        root = dict(root)
        reddit_emojis_root = root.pop('snoomojis')
        full_id36: str
        full_id36, subreddit_emojis_root = root.popitem()
        _, _, subreddit_id36 = full_id36.partition('_')
        subreddit_emojis = {name: load_flair_emoji(d, name) for name, d in subreddit_emojis_root.items()}
        reddit_emojis = {name: load_flair_emoji(d, name) for name, d in reddit_emojis_root.items()}
        all_emojis = {**subreddit_emojis, **reddit_emojis}
        return SubredditFlairEmojis(
            subreddit_emojis=subreddit_emojis,
            reddit_emojis=reddit_emojis,
            all_emojis=all_emojis,
            subreddit_id36=subreddit_id36,
        )

    class Create:
        def __init__(self, outer: FlairEmojiProcedures) -> None:
            self._client = outer._client

        async def __call__(self,
            sr: str,
            name: str,
            file: IO[bytes],
            *,
            filepath: Optional[str] = None,
            mod_only: bool = False,
            user_enabled: bool = True,
            post_enabled: bool = True,
            timeout: float = 1000,
        ) -> None:
            upload_lease = await self.upload(file, sr=sr, filepath=filepath, timeout=timeout)
            await self.add(sr, upload_lease.s3_object_key, name,
                    user_enabled=user_enabled,
                    post_enabled=post_enabled,
                    mod_only=mod_only)

        async def obtain_upload_lease(self,
            *,
            sr: str,
            filepath: str,
            mimetype: Optional[str] = None,
        ) -> FlairEmojiUploadLease:
            if mimetype is None:
                mimetype = guess_filename_mimetype(filepath)
            result = await self._client.request('POST', f'/api/v1/{sr}/emoji_asset_upload_s3',
                    data={'filepath': filepath, 'mimetype': mimetype})
            return load_flair_emoji_upload_lease(result)

        async def deposit_file(self,
            file: IO[bytes],
            upload_lease: FlairEmojiUploadLease,
            *,
            timeout: float = 1000,
        ) -> None:
            resp = await self._client.http.request('POST', upload_lease.endpoint,
                    data=upload_lease.fields, files={'file': file}, timeout=timeout)
            resp.ensure_successful_status()

        async def upload(self,
            file: IO[bytes],
            *,
            sr: str,
            filepath: Optional[str] = None,
            timeout: float = 1000,
        ) -> FlairEmojiUploadLease:
            if filepath is None:
                filepath = op.basename(getattr(file, 'name', ''))
                if not filepath:
                    raise ValueError("the `filepath` parameter must be explicitly specified if the file object has no `name` attribute.")
            upload_lease = await self.obtain_upload_lease(sr=sr, filepath=filepath)
            await self.deposit_file(file, upload_lease, timeout=timeout)
            return upload_lease

        async def add(self,
            sr: str,
            name: str,
            s3_object_key: str,
            *,
            mod_only: bool = False,
            user_enabled: bool = True,
            post_enabled: bool = True,
        ) -> None:
            data = {
                's3_key': s3_object_key,
                'name': name,
                'user_flair_allowed': '01'[user_enabled],
                'post_flair_allowed': '01'[post_enabled],
                'mod_flair_only': '01'[mod_only],
            }
            await self._client.request('POST', f'/api/v1/{sr}/emoji', data=data)

    create: cached_property[Create] = cached_property(Create)
    ("""
        Create a new flair emoji.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` name:
        :param `IO[bytes]` file:
        :param `Optional[str]` filepath:
        :param `bool` mod_only:
        :param `bool` post_enabled:
        :param `bool` user_enabled:
        :param `float` timeout:

        .. .RETURNS

        :rtype: `None`
        """)

    async def set_permissions(self,
        sr: str,
        name: str,
        *,
        mod_only: bool = False,
        user_enabled: bool = True,
        post_enabled: bool = True,
    ) -> None:
        """Change emoji permissions.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` name:
        :param `bool` mod_only:
        :param `bool` user_enabled:
        :param `bool` post_enabled:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You do not have permission.
            + `404`:
                The specified emoji doesn't exist.
            + `500`:
               - The specified subreddit does not exist.
               - The specified emoji name was empty.
        """
        data = {
            'name': name,
            'mod_flair_only': '01'[mod_only],
            'user_flair_allowed': '01'[user_enabled],
            'post_flair_allowed': '01'[post_enabled],
        }
        await self._client.request('POST', f'/api/v1/{sr}/emoji_permissions', data=data)

    async def delete(self, sr: str, name: str) -> None:
        """Delete a flair emoji.

        .. .PARAMETERS

        :param `str` sr:
        :param `str` name:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `400`:
               - The specified subreddit does not exist.
               - The specified emoji name was empty.
        """
        await self._client.request('DELETE', f'/api/v1/{sr}/emoji/{name}')

    async def enable_emojis_in_subreddit(self, sr: str) -> None:
        """Enable flair emojis in a subreddit.

        .. .PARAMETERS

        :param `str` sr:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You do not have permission.
            + `500`:
               - The specified subreddit does not exist.
               - The specified subreddit was empty.
        """
        data = {'subreddit': sr, 'enable': '1'}
        await self._client.request('POST', '/api/enable_emojis_in_sr', data=data)

    async def disable_emojis_in_subreddit(self, sr: str) -> None:
        """Disable flair emojis in a subreddit.

        .. .PARAMETERS

        :param `str` sr:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :(raises): Same as :meth:`.enable_emojis_in_subreddit`.
        """
        data = {'subreddit': sr, 'enable': '0'}
        await self._client.request('POST', '/api/enable_emojis_in_sr', data=data)




from __future__ import annotations

from dataclasses import dataclass

@dataclass
class ModReport:
    """A moderator report."""

    reason: str
    ("""
        Report reason provided by the moderator.
        """)
    agent_name: str
    ("""
        Name of the reporting moderator.

        Unknown what happens if the user is deleted.
        Is the value `[deleted]`, an empty string, or does the field in the
        underlying object not exist?
        If you have any information about this, please open an issue report at
        `<https://github.com/Pyprohly/redditwarp/issues>`_.
        """)

@dataclass
class UserReport:
    """An arbitrary user report."""

    reason: str
    ("""
        Report reason.
        """)
    count: int
    ("""
        Number of duplicate reports for this reason.
        """)
    snoozed: bool
    ("""
        Whether this report reason is snoozed.
        """)
    can_snooze: bool




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Mapping
if TYPE_CHECKING:
    from ..handler_ASYNC import Handler
    from ..send_params import SendParams
    from ..exchange import Exchange

from ..delegating_handler_ASYNC import DelegatingHandler


class ApplyParamsAndHeaders(DelegatingHandler):
    def __init__(self, handler: Handler, *,
            params: Optional[Mapping[str, str]] = None,
            headers: Optional[Mapping[str, str]] = None) -> None:
        super().__init__(handler)
        self.params: Mapping[str, str] = {} if params is None else params
        ("")
        self.headers: Mapping[str, str] = {} if headers is None else headers
        ("")

    async def _send(self, p: SendParams) -> Exchange:
        reqi = p.requisition
        reqi.params.update(self.params)
        reqi.headers.update(self.headers)
        return await super()._send(p)

class ApplyDefaultParamsAndHeaders(DelegatingHandler):
    def __init__(self, handler: Handler, *,
            params: Optional[Mapping[str, str]] = None,
            headers: Optional[Mapping[str, str]] = None) -> None:
        super().__init__(handler)
        self.params: Mapping[str, str] = {} if params is None else params
        ("")
        self.headers: Mapping[str, str] = {} if headers is None else headers
        ("")

    async def _send(self, p: SendParams) -> Exchange:
        reqi = p.requisition
        (pd := reqi.params).update({**self.params, **pd})
        (hd := reqi.headers).update({**self.headers, **hd})
        return await super()._send(p)


class ApplyParams(DelegatingHandler):
    def __init__(self, handler: Handler, params: Optional[Mapping[str, str]] = None) -> None:
        super().__init__(handler)
        self.params: Mapping[str, str] = {} if params is None else params
        ("")

    async def _send(self, p: SendParams) -> Exchange:
        reqi = p.requisition
        reqi.params.update(self.params)
        return await super()._send(p)

class ApplyDefaultParams(DelegatingHandler):
    def __init__(self, handler: Handler, params: Optional[Mapping[str, str]] = None) -> None:
        super().__init__(handler)
        self.params: Mapping[str, str] = {} if params is None else params
        ("")

    async def _send(self, p: SendParams) -> Exchange:
        reqi = p.requisition
        (pd := reqi.params).update({**self.params, **pd})
        return await super()._send(p)


class ApplyHeaders(DelegatingHandler):
    def __init__(self, handler: Handler, headers: Optional[Mapping[str, str]] = None) -> None:
        super().__init__(handler)
        self.headers: Mapping[str, str] = {} if headers is None else headers
        ("")

    async def _send(self, p: SendParams) -> Exchange:
        reqi = p.requisition
        reqi.headers.update(self.headers)
        return await super()._send(p)

class ApplyDefaultHeaders(DelegatingHandler):
    def __init__(self, handler: Handler, headers: Optional[Mapping[str, str]] = None) -> None:
        super().__init__(handler)
        self.headers: Mapping[str, str] = {} if headers is None else headers
        ("")

    async def _send(self, p: SendParams) -> Exchange:
        reqi = p.requisition
        (hd := reqi.headers).update({**self.headers, **hd})
        return await super()._send(p)




from .widget_ASYNC import Widget as Widget  # noqa: F401
from .text_area_ASYNC import TextAreaWidget as TextAreaWidget  # noqa: F401
from .button_ASYNC import ButtonWidget as ButtonWidget  # noqa: F401
from .image_ASYNC import ImageWidget as ImageWidget  # noqa: F401
from .community_list_ASYNC import CommunityListWidget as CommunityListWidget  # noqa: F401
from .calendar_ASYNC import CalendarWidget as CalendarWidget  # noqa: F401
from .post_flair_ASYNC import PostFlairWidget as PostFlairWidget  # noqa: F401
from .custom_css_ASYNC import CustomCSSWidget as CustomCSSWidget  # noqa: F401
from .community_details_ASYNC import CommunityDetailsWidget as CommunityDetailsWidget  # noqa: F401
from .moderator_list_ASYNC import ModeratorListWidget as ModeratorListWidget  # noqa: F401
from .rules_ASYNC import RulesWidget as RulesWidget  # noqa: F401

from .menu_bar_ASYNC import MenuBar as MenuBar  # noqa: F401

from .widget_list_ASYNC import WidgetList as WidgetList  # noqa: F401




from __future__ import annotations

from dataclasses import dataclass

@dataclass(repr=False, eq=False)
class Rule:
    description: str
    description_html: str
    short_name: str
    violation_reason: str
    created_ut: int




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence, Iterable
if TYPE_CHECKING:
    from ....client_ASYNC import Client

from ..listing.listing_async_paginator import ListingAsyncPaginator
from ....models.moderation_action_log_entry import ModerationActionLogEntry
from ....model_loaders.moderation_action_log_entry import load_moderation_action_log_entry

class ModerationActionLogAsyncPaginator(ListingAsyncPaginator[ModerationActionLogEntry]):
    def __init__(self,
        client: Client,
        url: str,
        *,
        limit: Optional[int] = 500,
        action: str = '',
        mod: str = '',
    ) -> None:
        cursor_extractor = lambda x: x['data']['id']
        super().__init__(client, url, limit=limit, cursor_extractor=cursor_extractor)
        self.client: Client = client
        ("")
        self.url: str = url
        ("")
        self.action: str = action
        ("")
        self.mod: str = mod
        ("")

    def _generate_params(self) -> Iterable[tuple[str, str]]:
        yield from super()._generate_params()
        if self.action:
            yield ('action', self.action)
        if self.mod:
            yield ('mod', self.mod)

    async def fetch(self) -> Sequence[ModerationActionLogEntry]:
        data = await self._fetch_data()
        return [load_moderation_action_log_entry(d['data']) for d in data['children']]




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..client_ASYNC import Client

from dataclasses import dataclass

from .message import (
    MailboxMessage as BaseMailboxMessage,
    ComposedMessage as BaseComposedMessage,
    CommentMessage as BaseCommentMessage,
)

@dataclass(repr=False, eq=False)
class MailboxMessage(BaseMailboxMessage):
    client: Client

@dataclass(repr=False, eq=False)
class ComposedMessage(MailboxMessage, BaseComposedMessage):
    async def reply(self, body: str) -> ComposedMessage:
        return await self.client.p.message.reply(self.id, body)

    async def mark_read(self) -> None:
        await self.client.p.message.mark_read(self.id)

    async def mark_unread(self) -> None:
        await self.client.p.message.mark_unread(self.id)

@dataclass(repr=False, eq=False)
class CommentMessage(MailboxMessage, BaseCommentMessage):
    async def mark_read(self) -> None:
        await self.client.p.message.mark_comment_read(self.comment.id)

    async def mark_unread(self) -> None:
        await self.client.p.message.mark_comment_unread(self.comment.id)




from __future__ import annotations
from typing import Sequence

from .submission_draft import SubmissionDraft, SubmissionDraftList as BaseSubmissionDraftList
from .subreddit_ASYNC import Subreddit

class SubmissionDraftList(BaseSubmissionDraftList):
    @property
    def subreddits(self) -> Sequence[Subreddit]:
        return self.__subreddits

    def __init__(self,
        drafts: Sequence[SubmissionDraft],
        subreddits: Sequence[Subreddit],
    ) -> None:
        super().__init__(drafts, subreddits)
        self.__subreddits: Sequence[Subreddit] = subreddits




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence, Iterable, Union, Mapping
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.comment_ASYNC import Comment
    from ...types import JSON_ro

import json

from ...model_loaders.comment_ASYNC import load_comment
from ...util.base_conversion import to_base36
from ...iterators.chunking import chunked
from ...iterators.call_chunk_chaining_async_iterator import CallChunkChainingAsyncIterator
from ...iterators.async_call_chunk import AsyncCallChunk
from .fetch_ASYNC import Fetch
from .get_ASYNC import Get

class CommentProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client
        self.fetch: Fetch = Fetch(self, client)
        ("""
            Fetch a comment.

            .. .PARAMETERS

            :param `int` idn:
                Comment ID.

            .. .RETURNS

            :rtype: :class:`~.models.comment_ASYNC.Comment`

            .. .RAISES

            :raises redditwarp.exceptions.NoResultException:
                The target was not found.
            """)
        self.get: Get = Get(client)
        ("""
            Get a comment.

            .. .PARAMETERS

            :param `int` idn:
                Comment ID.

            .. .RETURNS

            :rtype: `Optional`\\[:class:`~.models.comment_ASYNC.Comment`]
            """)

    def bulk_fetch(self, ids: Iterable[int]) -> CallChunkChainingAsyncIterator[Comment]:
        """Bulk fetch comments.

        Any ID not found will be ignored.

        .. .PARAMETERS

        :param `Iterable[int]` ids:
            Comment IDs.

        .. .RETURNS

        :rtype: :class:`~.iterators.call_chunk_chaining_async_iterator.CallChunkChainingAsyncIterator`\\[:class:`~.models.comment_ASYNC.Comment`]
        """
        async def mass_fetch(ids: Sequence[int]) -> Sequence[Comment]:
            id36s = map(to_base36, ids)
            full_id36s = map('t1_'.__add__, id36s)
            ids_str = ','.join(full_id36s)
            root = await self._client.request('GET', '/api/info', params={'id': ids_str})
            return [load_comment(i['data'], self._client) for i in root['data']['children']]

        return CallChunkChainingAsyncIterator(AsyncCallChunk(mass_fetch, idfs) for idfs in chunked(ids, 100))

    async def reply(self, idn: int, body: Union[str, Mapping[str, JSON_ro]]) -> Comment:
        """Reply to a comment.

        .. .PARAMETERS

        :param `int` idn:
        :param body:
            Either markdown or richtext.
        :type body: `Union`\\[`str`, `Mapping`\\[`str`, :class:`~.types.JSON_ro`]]

        .. .RETURNS

        :rtype: :class:`~.models.comment_ASYNC.Comment`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `RATELIMIT`:
                Content creation cooldown.
            + `NO_TEXT`:
                The specified `body` was empty.
            + `TOO_OLD`:
                The subreddit has archiving enabled and the target comment
                is archived (older than 6 months).
            + `THREAD_LOCKED`:
                The target comment submission is locked and you are not a
                moderator of the subreddit.
            + `DELETED_COMMENT`:
                The target comment was deleted.
            + `SOMETHING_IS_BROKEN`:
                The author of the target comment has blocked you.
            + `USER_BLOCKED`:
                The author of the target comment is a user you have blocked.
            + `SUBREDDIT_OUTBOUND_LINKING_DISALLOWED`:
                Some subreddits prevent you from linking to other subreddits. E.g.,
                writing 'r/funny' in 'r/formuladank'. It is not known what setting
                controls this.
            + `SUBREDDIT_LINKING_DISALLOWED`:
                Some subreddits cannot be linked to at all. E.g., 'r/chonglangTV'.
                It is unknown why.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                The target comment does not exist.
            + `500`:
                The target comment is from a quarantined subreddit that the
                current user has not opted in to.
        """
        def g() -> Iterable[tuple[str, str]]:
            yield ('thing_id', 't1_' + to_base36(idn))
            yield ('return_rtjson', '1')
            if isinstance(body, str):
                yield ('text', body)
            else:
                yield ('richtext_json', json.dumps(body))

        result = await self._client.request('POST', '/api/comment', files=dict(g()))
        return load_comment(result, self._client)

    async def edit_body(self, idn: int, body: Union[str, Mapping[str, JSON_ro]]) -> Comment:
        """Edit the body text of a comment.

        The target entity (with the new body text) is returned.

        .. .PARAMETERS

        :param `int` idn:
        :param body:
            Either markdown or richtext.
        :type body: `Union`\\[`str`, `Mapping`\\[`str`, :class:`~.types.JSON_ro`]]

        .. .RETURNS

        :rtype: :class:`~.models.comment_ASYNC.Comment`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_THING_ID`:
                The target comment does not exist.
        """
        def g() -> Iterable[tuple[str, str]]:
            yield ('thing_id', 't1_' + to_base36(idn))
            yield ('return_rtjson', '1')
            if isinstance(body, str):
                yield ('text', body)
            else:
                yield ('richtext_json', json.dumps(body))

        result = await self._client.request('POST', '/api/editusertext', files=dict(g()))
        return load_comment(result, self._client)

    async def delete(self, idn: int) -> None:
        """Delete a comment.

        If the target doesn't exist or isn't valid, nothing happens.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        data = {'id': 't1_' + to_base36(idn)}
        await self._client.request('POST', '/api/del', data=data)

    async def lock(self, idn: int) -> None:
        """Lock a comment.

        Nothing happens if the target is already locked.

        .. hint::

           Locking prevents a submission/comment from receiving new comments.
           A locked submission is unable to receive any new comments.
           Locking a comment only stops direct comments, but
           existing child comments can still receive replies.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                The target doesn't exist or you don't have permission to lock it.
        """
        data = {'id': 't1_' + to_base36(idn)}
        await self._client.request('POST', '/api/lock', data=data)

    async def unlock(self, idn: int) -> None:
        """Unlock a comment.

        Behaves similarly to :meth:`.lock`.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises:
            Similar to :meth:`.lock`.
        """
        data = {'id': 't1_' + to_base36(idn)}
        await self._client.request('POST', '/api/unlock', data=data)

    async def vote(self, idn: int, direction: int) -> None:
        """Cast a vote on a comment.

        .. .PARAMETERS

        :param `int` idn:
        :param `int` direction:
            Either: `1` (upvote), `0` unvote, `-1` downvote.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `404`:
                The target could not be found.
        """
        data = {
            'id': 't1_' + to_base36(idn),
            'dir': str(direction),
        }
        await self._client.request('POST', '/api/vote', data=data)

    async def save(self, idn: int, category: Optional[str] = None) -> None:
        """Save a comment.

        .. .PARAMETERS

        :param `int` idn:
        :param `Optional[str]` category:
            A category/label.

            Requires Reddit Premium. Ignored if no Reddit Premium.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `404`:
                The category name specified was invalid.
        """
        data = {
            'id': 't1_' + to_base36(idn),
        }
        if category is not None:
            data['category'] = category
        await self._client.request('POST', '/api/save', data=data)

    async def unsave(self, idn: int) -> None:
        """Save a comment.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        data = {'id': 't1_' + to_base36(idn)}
        await self._client.request('POST', '/api/unsave', data=data)

    async def distinguish(self, idn: int) -> Comment:
        """Distinguish a comment.

        .. hint::
           Distinguishing decoratates the author's name by
           giving it a different color and putting a sigil beside it.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :returns: The target comment.
        :rtype: :class:`~.models.comment_ASYNC.Comment`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You do not have permission to distinguish the target.
            + `404`:
                The target could not be found.
        """
        data = {
            'id': 't1_' + to_base36(idn),
            'how': 'yes',
        }
        root = await self._client.request('POST', '/api/distinguish', data=data)
        return load_comment(root['json']['data']['things'][0]['data'], self._client)

    async def distinguish_and_sticky(self, idn: int) -> Comment:
        """Distinguish and sticky a comment.

        Only one comment may be stickied at a time. Attempting to sticky a comment
        when there is already a stickied comment will override that stickied comment.
        Only top-level comments may be stickied.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :returns: The target comment.
        :rtype: :class:`~.models.comment_ASYNC.Comment`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You do not have permission to distinguish the target.
            + `404`:
                The target could not be found.
        """
        data = {
            'id': 't1_' + to_base36(idn),
            'how': 'yes',
            'sticky': '1',
        }
        root = await self._client.request('POST', '/api/distinguish', data=data)
        return load_comment(root['json']['data']['things'][0]['data'], self._client)

    async def undistinguish(self, idn: int) -> Comment:
        """Undistinguish a comment.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :returns: The target comment.
        :rtype: :class:`~.models.comment_ASYNC.Comment`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You do not have permission.
            + `404`:
                The target could not be found.
        """
        data = {
            'id': 't1_' + to_base36(idn),
            'how': 'no',
        }
        root = await self._client.request('POST', '/api/distinguish', data=data)
        return load_comment(root['json']['data']['things'][0]['data'], self._client)

    async def enable_reply_notifications(self, idn: int) -> None:
        """Enable inbox reply notifications for a comment.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        data = {
            'id': 't1_' + to_base36(idn),
            'state': '1'
        }
        await self._client.request('POST', '/api/sendreplies', data=data)

    async def disable_reply_notifications(self, idn: int) -> None:
        """Disable inbox reply notifications for a comment.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        data = {
            'id': 't1_' + to_base36(idn),
            'state': '0'
        }
        await self._client.request('POST', '/api/sendreplies', data=data)

    async def approve(self, idn: int) -> None:
        """Approve a comment.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               You do not have permission.

               - The target specified does not exist.
               - The target belongs to a subreddit you do not have permission over.
        """
        data = {'id': 't1_' + to_base36(idn)}
        await self._client.request('POST', '/api/approve', data=data)

    async def remove(self, idn: int) -> None:
        """Remove a comment.

        This is a moderator action.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               You do not have permission.

               - The target specified does not exist.
               - The target belongs to a subreddit you do not have permission over.
        """
        data = {
            'id': 't1_' + to_base36(idn),
            'spam': '0',
        }
        await self._client.request('POST', '/api/remove', data=data)

    async def remove_spam(self, idn: int) -> None:
        """Remove as spam.

        See :meth:`.remove`.
        """
        data = {
            'id': 't1_' + to_base36(idn),
            'spam': '1',
        }
        await self._client.request('POST', '/api/remove', data=data)

    async def ignore_reports(self, idn: int) -> None:
        """Ignore reports on a comment.

        .. hint::
           If you ignore reports, you won't receive notifications and the
           ignored thing will be absent from moderation listings.

        Nothing happens if the target is already ignored.

        .. .PARAMETERS

        :param `int` idn:
            Comment ID.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - The target specified does not exist.
               - The target belongs to a subreddit you do not have permission over.
        """
        await self._client.request('POST', '/api/ignore_reports', data={'id': 't1_' + to_base36(idn)})

    async def unignore_reports(self, idn: int) -> None:
        """Unignore reports on a comment.

        Nothing happens if the target is already unignored.

        .. .PARAMETERS

        :param `int` idn:
            Comment ID.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - The target specified does not exist.
               - The target belongs to a subreddit you do not have permission over.
        """
        await self._client.request('POST', '/api/unignore_reports', data={'id': 't1_' + to_base36(idn)})

    async def snooze_reports(self, idn: int, reason: str) -> None:
        """Ignore a custom report reason in a subreddit for 7 days.

        .. .PARAMETERS

        :param `int` idn:
            Comment ID.
        :param `str` reason:
            The custom report reason to snooze.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - The target specified does not exist.
               - The target belongs to a subreddit you do not have permission over.
        """
        data = {'id': 't1_' + to_base36(idn), 'reason': reason}
        await self._client.request('POST', '/api/snooze_reports', data=data)

    async def unsnooze_reports(self, idn: int, reason: str) -> None:
        """Unsnooze a custom report.

        See :meth:`.snooze_reports`.
        """
        data = {'id': 't1_' + to_base36(idn), 'reason': reason}
        await self._client.request('POST', '/api/unsnooze_reports', data=data)

    async def apply_removal_reason(self,
            idn: int,
            reason_id: Optional[str] = None,
            note: Optional[str] = None) -> None:
        """Set a removal reason on a removed comment.

        If the target is not a removed comment, nothing happens.

        .. .PARAMETERS

        :param `int` idn:
            Comment ID.
        :param `Optional[int]` reason_id:
            A removal reason ID.
        :param `Optional[str]` note:
            A note.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `JSON_MISSING_KEY`:
                Empty strings or null values were specified for both
                `reason_id` and `note` at the same time.
            + `NO_THING_ID`:
                The given target ID is not valid.
            + `INVALID_ID`:
                The reason ID is invalid.
            + `MUST_BE_PRESENT`:
                The subreddit specified does not exist.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - The target specified does not belong to a subreddit you moderate.
        """
        target = 't1_' + to_base36(idn)
        json_data = {'item_ids': [target], 'reason_id': reason_id, 'mod_note': note}
        await self._client.request('POST', '/api/v1/modactions/removal_reasons', json=json_data)

    async def send_removal_comment(self,
            idn: int,
            title: str,
            message: str,
            *,
            exposed: bool = False,
            locked: bool = False) -> Comment:
        """Send a removal comment.

        Sends a removal reason comment to a user for a removed comment of theirs.

        This action can be performed multiple times.
        (The UI does not normally let you do this.)

        Unlike :meth:`.apply_removal_reason`, the target you specify must be a
        removed item otherwise an `INVALID_ID` API error is produced.

        .. .PARAMETERS

        :param `int` idn:
            ID of a removed comment.
        :param `str` title:
            A title.

            This is ultimately unused for removal comments, but a non-empty
            string must be specified or you'll get a `NO_TEXT` API error.

            The UI sends the title of the selected removal reason.
        :param `str` message:
            A message for the comment body.

            Can be an empty string. This is interesting because you can't
            normally create comments with empty bodies.
        :param `bool` exposed:
            If false (default), the comment will be created by a special
            moderator named `u/{subreddit}_ModTeam`.

            If true, the comment is created by the current user.
        :param `bool` locked:
            Lock the newly created comment.

        .. .RETURNS

        :returns: The newly created comment.
        :rtype: :class:`~.models.comment_ASYNC.Comment`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_TEXT`:
                The value for the `title` parameter was empty.
            + `INVALID_ID`:
               - The target specified doesn't exist or is invalid.
               - The target specified is not a removed item.

            + `MUST_BE_PRESENT`:
                The subreddit specified does not exist.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - The target specified does not belong to a subreddit you moderate.
        """
        target = 't1_' + to_base36(idn)
        json_data = {
            'type': 'public' + ('' if exposed else '_as_subreddit'),
            'item_id': [target],
            'title': title,
            'message': message,
            'lock_comment': '01'[locked],
        }
        root = await self._client.request('POST', '/api/v1/modactions/removal_comment_message', json=json_data)
        return load_comment(root, self._client)

    async def send_removal_message(self,
            idn: int,
            title: str,
            message: str,
            *,
            exposed: bool = False) -> None:
        """Send a removal message.

        Behaves similarly to :meth:`.send_removal_comment`.

        .. .PARAMETERS

        :param `int` idn:
            ID of a removed comment.
        :param `str` title:
            A title.

            A non-empty string must be specified or you'll get a `NO_TEXT` API error.

            The UI sends the title of the selected removal reason.
        :param `str` message:
            A message for the comment body.

            Can be an empty string.
        :param `bool` exposed:
            If false (default), the comment will be send on behalf of the subreddit.

            If true, the comment is sent by the current user.

        .. .RETURNS

        :returns: `None`

        .. .RAISES

        :(raises): See :meth:`.send_removal_comment`.
        """
        target = 't1_' + to_base36(idn)
        json_data = {
            'type': 'private' + ('_exposed' if exposed else ''),
            'item_id': [target],
            'title': title,
            'message': message,
        }
        await self._client.request('POST', '/api/v1/modactions/removal_comment_message', json=json_data)




from __future__ import annotations

from ..listing.mixins.subreddit_detail_ASYNC import SubredditDetail
from ..listing.subreddit_listing_async_paginator import SubredditListingAsyncPaginator
from ....models.subreddit_ASYNC import Subreddit

class UserSubredditListingAsyncPaginator(
    SubredditDetail[Subreddit],
    SubredditListingAsyncPaginator,
): pass
